import pandas as pd
import numpy as np

from pysesd.sesd import SESD


def get_ts():
    """
    The get_ts function creates a random time series with two outliers.

    :return: A time series
    """
    dates = pd.date_range(start='2023-01-01', end='2023-06-19', freq='D')
    values = np.random.random(len(dates))
    values[42] = 10
    values[24] = 20

    ts = pd.Series(data=values, index=dates)
    return ts


def run():
    """
    The run function is the main function of this module.
    It will be called by the user to run the SESD algorithm on a time series.


    :return: The list of outliers
    """
    ts = get_ts()
    sesd = SESD(alpha=0.05, hybrid=False, max_outliers=2)
    outliers = sesd.fit(ts)
    print(f"Got {len(outliers)} in {len(ts)} points, anomaly index: {outliers}")
    sesd.plot(save=True, fig_dir="../figures", fig_name="simple.png")


if __name__ == '__main__':
    run()



from typing import Optional, List

import matplotlib.pyplot as plt
import pandas as pd
import logging
import pathlib

from pysesd.util import esd_test, calculate_stl_residual

logger = logging.getLogger(__name__)


class SESD:
    def __init__(self,
                 alpha: float = 0.05,
                 hybrid: bool = False,
                 period: Optional[int] = None,
                 max_outliers: Optional[int] = None,
                 ):
        """
        The __init__ function is called when the class is instantiated.
        It sets up the attributes of an instance of a class.

        :param alpha: Set the significance level
        :param hybrid: Determine whether the algorithm is hybrid or not
        :param period: Specify the period of the time series
        :param max_outliers: Specify the maximum number of outliers that can be detected
        :param : Set the alpha value for the test
        :return: None
        """
        self.alpha = alpha
        self.hybrid = hybrid
        self.period = period
        self.max_outliers = max_outliers
        # for plot
        self.outliers = None
        self.ts = None

    def fit(self, ts: pd.Series) -> List[int]:
        # TODO: ts index/NA check
        """
        The fit function takes a time series and returns the indices of outliers.

        :param ts: pd.Series: Pass in the time series data
        :return: The list of outliers
        """
        # data check: ts
        if not isinstance(ts.index, pd.DatetimeIndex):
            try:
                ts.index = pd.DatetimeIndex(ts.index)
            except Exception as e:
                msg = f"The ts data's type is {type(ts.index)} which can't be converted to `pd.DatetimeIndex`: {e}"
                logger.error(msg)
                raise ValueError(msg)

        self.ts = ts
        # parameter check: max_outliers
        n = len(ts)
        if self.max_outliers >= n // 2:
            raise ValueError(f"max_outliers = {self.max_outliers} >= {n // 2}({n} // 2)")
        # parameter check: period
        if self.period is None:
            self.period = self.__infer_period(ts.index)
        # calc STL residual
        residuals = calculate_stl_residual(ts.values, period=self.period)
        self.outliers = esd_test(residuals, self.alpha, self.hybrid, self.max_outliers)
        return self.outliers

    @staticmethod
    def __infer_period(datetime_index: pd.DatetimeIndex) -> int:
        """
        Reference: https://github.com/twitter/AnomalyDetection/blob/1f5deaa1609f8f1964c1e905c7a8ad2d1d0dc718/R/ts_anom_detection.R#LL168C1-L173C27  # NOQA
        :param datetime_index: The time series' index
        :return: The inferred period
        """
        infer_freq = pd.infer_freq(datetime_index)
        if infer_freq == "S":  # second -> one day
            period = 86400
        elif infer_freq == "T" or infer_freq == "min":
            period = int(86400 / 60)  # minute -> one day
        elif infer_freq == "H":
            period = int(86400 / 3600)  # hour -> one day
        elif infer_freq == "D":  # day -> one week
            period = 7
        else:  # TODO: more elegant handle method
            period = int(0.1 * len(datetime_index))
        return period

    def plot(self, save: bool = True, fig_name: str = "sesd.png", fig_dir: str = "../figures"):
        """
        The plot function plots the time series and the anomaly points.

        :param self: Represent the instance of the class
        :param save: Determine whether the plot should be saved or not
        :param fig_name: Name the file that is saved
        :param fig_dir: Specify the directory where the figure will be saved
        :return: None
        """
        timestamps = self.ts.index
        values = self.ts.values

        plt.style.use('ggplot')
        _, ax = plt.subplots(figsize=(10, 6))

        # Create a plot for the time series
        ax.plot(timestamps, values, color="lightcoral")

        # Plot the anomaly points
        for point_index in self.outliers:
            ax.plot(timestamps[point_index], values[point_index], color="cyan", marker="*", markersize=10)

        # Set the title and axis labels
        ax.set_title('S-ESH')
        ax.set_xlabel('Date')
        ax.set_ylabel('Value')
        # Add a legend
        ax.legend(['Time Series', 'Anomaly Points'])
        plt.xticks(rotation=60)
        plt.tight_layout()
        if save:
            pathlib.Path(fig_dir).mkdir(parents=True, exist_ok=True)
            fig_path = f"{fig_dir}/{fig_name}"
            plt.savefig(fig_path, dpi=300)
        plt.show()



import pandas as pd

from pysesd.sesd import SESD


def get_ts():
    """
    The get_ts function creates a random time series with two outliers.

    :return: A time series
    """
    ts = pd.read_csv("../dataset/twitter_raw_data.csv", parse_dates=["timestamp"], index_col=["timestamp"])
    ts = ts.iloc[:, 0]
    ts.index = pd.DatetimeIndex(ts.index)
    return ts


def run():
    """
    The run function is the main function of this module.
    It will be called by the user to run the SESD algorithm on a time series.

    :return: The list of outliers
    """
    ts = get_ts()
    sesd = SESD(alpha=0.05, hybrid=True, max_outliers=int(len(ts) * 0.02))
    outliers = sesd.fit(ts)
    print(f"Got {len(outliers)} in {len(ts)} points, anomaly index: {outliers}")
    sesd.plot(save=True, fig_dir="../figures", fig_name="twitter.png")


if __name__ == '__main__':
    run()



from typing import List, Optional

import numpy as np
import scipy
from statsmodels.tsa.seasonal import STL


def esd_test(data: np.array,
             alpha: float = 0.05,
             hybrid: bool = False,
             max_outliers: Optional[int] = None) -> List[int]:
    """
    Perform the Extreme Studentized Deviate (ESD) test to detect potential outliers in the data.

    :param data: The input data to test for potential outliers.
    :param alpha: The significance level for the test. Default is 0.05.
    :param hybrid: Use hybrid method or not. Default is False.
    :param max_outliers: The maximum number of outliers to search for. Default is None,
    which sets max_outliers to len(data) // 2.
    :return: The indices of the detected outliers in the input data.
    """
    if max_outliers is None:
        max_outliers = len(data) // 2

    data = np.ma.array(data)
    outliers = []

    for _ in range(max_outliers):
        test_statistic, test_statistic_idx = calc_test_statistic(data, hybrid)

        # Compute the critical value
        n = len(data) - data.mask.sum()
        t_value = scipy.stats.t.ppf(1 - alpha / (2 * n), n - 2)
        critical_value = (n - 1) * t_value / np.sqrt(np.square((n - 2)) + (n - 2) * np.square(t_value))

        # Compare the test statistic with the critical value
        if test_statistic > critical_value:
            outliers.append(test_statistic_idx)
            data[test_statistic_idx] = np.ma.masked
        else:
            break

    return outliers


def calc_test_statistic(data: np.ma.array, hybrid: bool = False) -> (float, int):
    """
    The calc_test_statistic function calculates the test statistic for a given data set.

    :param data: Pass the data to be tested
    :param hybrid: Determine whether to use the mean or median(hybrid) for the location parameter
    :return: A tuple of the test statistic and its index
    """
    if hybrid:
        loc_value = np.mean(data)
        scale_value = np.std(data)
    else:
        loc_value = np.median(data)
        scale_value = np.median(np.abs(data - loc_value))

    abs_dev_value = np.abs(data - loc_value)
    max_dev_value_index = np.argmax(abs_dev_value)
    max_dev_value = abs_dev_value[max_dev_value_index]
    test_statistic = max_dev_value / scale_value
    test_statistic_index = max_dev_value_index

    return test_statistic, test_statistic_index


def calculate_stl_residual(data: np.ndarray, period: Optional[int] = None) -> np.ndarray:
    """
    The calculate_stl_residual function takes a time series and returns the STL residuals.

    :param data: Pass the data to be used in the stl decomposition
    :param period: Specify the period of the time series
    :return: The residuals of the stl decomposition
    """
    stl = STL(data, period=period, robust=True)
    decomposition = stl.fit()
    residual = data - decomposition.seasonal - np.median(data)
    return residual
